package teamcity.crowd.integration.sync;

import com.atlassian.crowd.exception.ApplicationPermissionException;
import com.atlassian.crowd.exception.InvalidAuthenticationException;
import com.atlassian.crowd.exception.OperationFailedException;
import com.atlassian.crowd.exception.UserNotFoundException;
import com.atlassian.crowd.model.group.Membership;
import com.atlassian.crowd.service.client.CrowdClient;
import com.google.common.base.Function;
import com.google.common.collect.Collections2;
import com.google.common.collect.Sets;
import jetbrains.buildServer.groups.SUserGroup;
import jetbrains.buildServer.groups.UserGroupManager;
import jetbrains.buildServer.users.SUser;
import jetbrains.buildServer.users.User;
import jetbrains.buildServer.users.UserModel;
import teamcity.crowd.integration.*;
import teamcity.crowd.integration.utils.*;

import javax.annotation.Nullable;
import java.util.*;

public class CrowdGroupsSync implements CrowdSync {
    private final UsersAndGroupsListsToMapsConverter converter = new UsersAndGroupsListsToMapsConverter();

    private UserGroupManager userGroupManager;
    private CrowdClient crowdClient;
    private UserModel userModel;
    private LoggerProvider loggerProvider;

    public CrowdGroupsSync(UserGroupManager userGroupManager, CrowdClient crowdClient, UserModel userModel, LoggerProvider loggerProvider) {
        this.userGroupManager = userGroupManager;
        this.crowdClient = crowdClient;
        this.userModel = userModel;
        this.loggerProvider = loggerProvider;
    }

    private void syncGroups() {
        try {

            final Iterable<Membership> memberships = crowdClient.getMemberships();
            final Map<String, Set<String>> crowdMembersPerGroup = converter.membershipsAsMap(memberships);

            final Collection<SUserGroup> userGroups = userGroupManager.getUserGroups();
            final Map<String, SUserGroup> teamCityUsersPerGroup = converter.userGroupsAsMap(userGroups);

            Map<String, User> allTeamCityUsers = converter.allTeamCityUsersAsMap(userGroupManager.getAllUsersGroup().getAllUsers());

            logSomeAwesomeStatisticsFirst(crowdMembersPerGroup, teamCityUsersPerGroup, allTeamCityUsers);

            final Set<String> newCrowdGroups = Sets.newHashSet(crowdMembersPerGroup.keySet());
            newCrowdGroups.removeAll(teamCityUsersPerGroup.keySet());

            final Set<String> deletedCrowdGroups = getGroupsForDeletion(crowdMembersPerGroup, teamCityUsersPerGroup);

            final Set<String> groupsToBeUpdated = getExistingGroupsThatNeedUpdate(crowdMembersPerGroup, teamCityUsersPerGroup);

            addAllNewCrowdGroups(newCrowdGroups, crowdMembersPerGroup, allTeamCityUsers);
            removeDeletedCrowdGroups(deletedCrowdGroups, teamCityUsersPerGroup);

            // Need to fetch all the TC user as there could be more
            allTeamCityUsers = converter.allTeamCityUsersAsMap(userGroupManager.getAllUsersGroup().getAllUsers());
            updateGroups(groupsToBeUpdated, crowdMembersPerGroup, teamCityUsersPerGroup, allTeamCityUsers);

        } catch (OperationFailedException e) {
            loggerProvider.getServerLogger().error("Problems when trying to sync the Groups", e);
        } catch (ApplicationPermissionException e) {
            loggerProvider.getServerLogger().error("It doesn't look like TeamCity Crowd Integration plugin has the right Crowd permissions setup.", e);
        } catch (InvalidAuthenticationException e) {
            loggerProvider.getServerLogger().error("It doesn't look like TeamCity Crowd Integration plugin has the right Crowd permissions setup.", e);

        }
    }

    /**
     * The easiest thing will be to remove all users and add them again. Simples!
     */
    private void updateGroups(Set<String> groupsToBeUpdated, Map<String, Set<String>> crowdMembersPerGroup, Map<String, SUserGroup> teamCityUsersPerGroup, Map<String, User> allTeamCityUsers) {
        loggerProvider.getServerLogger().info("There is a [" + groupsToBeUpdated.size() + "] groups that need update");
        for (String groupName : groupsToBeUpdated) {
            final SUserGroup userGroup = teamCityUsersPerGroup.get(groupName);
            final List<User> usersToRemove = userGroup.getAllUsers();
            for (User user : usersToRemove) {
                userGroup.removeUser(user);
            }
            for (String usersInGroup : crowdMembersPerGroup.get(groupName)) {
                addUserToGroup(allTeamCityUsers, userGroup, usersInGroup);
            }
            loggerProvider.getServerLogger().info("Updated group [" + groupName + "] with new membership");
        }
    }

    private Set<String> getExistingGroupsThatNeedUpdate(Map<String, Set<String>> crowdMembersPerGroup, Map<String, SUserGroup> teamCityUsersPerGroup) {
        final Sets.SetView<String> groupsInBothSystems = Sets.intersection(crowdMembersPerGroup.keySet(), teamCityUsersPerGroup.keySet());
        Set<String> groupsThatNeedUpdate = new HashSet<String>();
        for (String groupName : groupsInBothSystems) {
            if (usersChangedInGroup(crowdMembersPerGroup.get(groupName), teamCityUsersPerGroup.get(groupName))) {
                groupsThatNeedUpdate.add(groupName);
            }
        }
        return groupsThatNeedUpdate;
    }

    private boolean usersChangedInGroup(Set<String> usersInCrowdGroup, SUserGroup userGroup) {
        final Set<String> usersInTeamCityGroup = Sets.newHashSet(Collections2.transform(userGroup.getAllUsers(), new Function<User, String>() {
            @Override
            public String apply(@Nullable User input) {
                return input.getUsername();
            }
        }));
        return !Sets.difference(usersInCrowdGroup, usersInTeamCityGroup).isEmpty() ||
                !Sets.difference(usersInTeamCityGroup, usersInCrowdGroup).isEmpty();
    }

    private Set<String> getGroupsForDeletion(Map<String, Set<String>> crowdMembersPerGroup, Map<String, SUserGroup> teamCityUsersPerGroup) {
        final Set<String> deletedCrowdGroups = Sets.newHashSet(teamCityUsersPerGroup.keySet());
        deletedCrowdGroups.removeAll(crowdMembersPerGroup.keySet());
        if (deletedCrowdGroups.isEmpty()) {
            loggerProvider.getServerLogger().info("There is no groups for Deletion!");
        } else {
            loggerProvider.getServerLogger().info("Groups for deletion = " + CollectionToString.flattenCollection(deletedCrowdGroups, ", "));
        }
        return deletedCrowdGroups;
    }

    private void removeDeletedCrowdGroups(Set<String> deletedCrowdGroups, Map<String, SUserGroup> teamCityUsersPerGroup) {
        for (String deletedCrowdGroup : deletedCrowdGroups) {
            final SUserGroup userGroup = teamCityUsersPerGroup.get(deletedCrowdGroup);
            userGroupManager.deleteUserGroup(userGroup);

            loggerProvider.getServerLogger().info("Group [" + deletedCrowdGroup + "] deleted in TeamCity");
        }

    }

    private void addAllNewCrowdGroups(Set<String> newCrowdGroups, Map<String, Set<String>> crowdMembersPerGroup, Map<String, User> allTeamCityUsers) {
        for (String newCrowdGroup : newCrowdGroups) {
            final SUserGroup userGroup = userGroupManager.createUserGroup(createGroupKey(newCrowdGroup), newCrowdGroup, "Created by Crowd Integration Plugin Sync");
            final Set<String> usersToAddToGroup = crowdMembersPerGroup.get(newCrowdGroup);

            for (String user : usersToAddToGroup) {
                addUserToGroup(allTeamCityUsers, userGroup, user);
            }

            loggerProvider.getServerLogger().info("New group [" + newCrowdGroup + "] added");
        }
    }

    private void addUserToGroup(Map<String, User> allTeamCityUsers, SUserGroup userGroup, String user) {
        if (allTeamCityUsers.containsKey(user)) {
            userGroup.addUser(allTeamCityUsers.get(user));
        } else {
            Maybe<SUser> newCreatedUser = createNewUser(user);
            if (newCreatedUser.hasValue())
                userGroup.addUser(newCreatedUser.getValue());
            else {
                final String errorMessage = String.format("Something is not right. User [%s] is assign to a group [%s] in crowd but cannot be found in there",
                        user, userGroup.getName());
                loggerProvider.getServerLogger().error(errorMessage);
            }
        }
    }

    private String createGroupKey(String newCrowdGroup) {
        return "" + newCrowdGroup.hashCode();
    }

    private Maybe<SUser> createNewUser(String user) {
        try {

            final com.atlassian.crowd.model.user.User userFromCrowd = crowdClient.getUser(user);
            final SUser account = userModel.createUserAccount(CrowdAuthenticationScheme.REALM, user);
            account.updateUserAccount(user, userFromCrowd.getName(), userFromCrowd.getEmailAddress());

            loggerProvider.getServerLogger().info("New user [" + user + "] added");

            return new Something<SUser>(account);
        } catch (UserNotFoundException e) {
            loggerProvider.getServerLogger().error(e);
        } catch (OperationFailedException e) {
            loggerProvider.getServerLogger().error(e);
        } catch (ApplicationPermissionException e) {
            loggerProvider.getServerLogger().error(e);
        } catch (InvalidAuthenticationException e) {
            loggerProvider.getServerLogger().error(e);
        }
        return new Nothing<SUser>();
    }

    @Override
    public void run() {
        try {
            loggerProvider.getServerLogger().warn("--------------- SYNC ZE GROUPS NOW ------------------ ");
            syncGroups();
            loggerProvider.getServerLogger().warn("--------------- SYNC GROUPS COMPLETED WITH NO ERRORS------------------ ");
        } catch (Exception ex) {
            loggerProvider.getServerLogger().error("----------------- Massive problem when trying to sync the Crowd groups.", ex);
        }
    }


    private void logSomeAwesomeStatisticsFirst(Map<String, Set<String>> crowdMembersPerGroup, Map<String, SUserGroup> teamCityUsersPerGroup, Map<String, User> allTeamCityUsers) {
        loggerProvider.getServerLogger().info("------------------------------------- Some facts before starting to sync ---------------------------------------------");
        String stats = String.format("There is [%d] groups in Crowd that have users assosiated. There is [%d] groups registered in TeamCity at the moment. There is [%d] all users in TeamCity.",
                crowdMembersPerGroup.size(), teamCityUsersPerGroup.size(), allTeamCityUsers.size());
        loggerProvider.getServerLogger().info(stats);
        loggerProvider.getServerLogger().info("---------------------------------------------------------------------------------------------------------------------");
    }

}
