package teamcity.crowd.integration.sync;

import com.atlassian.crowd.exception.ApplicationPermissionException;
import com.atlassian.crowd.exception.InvalidAuthenticationException;
import com.atlassian.crowd.exception.OperationFailedException;
import com.atlassian.crowd.exception.UserNotFoundException;
import com.atlassian.crowd.service.client.CrowdClient;
import jetbrains.buildServer.groups.UserGroupManager;
import jetbrains.buildServer.users.SUser;
import jetbrains.buildServer.users.User;
import jetbrains.buildServer.users.UserModel;
import org.apache.commons.lang.StringUtils;
import teamcity.crowd.integration.utils.LoggerProvider;

import java.util.Map;

public class CrowdUsersSync implements CrowdSync {
    private final UsersAndGroupsListsToMapsConverter converter = new UsersAndGroupsListsToMapsConverter();

    private UserGroupManager userGroupManager;
    private final LoggerProvider loggerProvider;
    private final CrowdClient crowdClient;
    private final UserModel userModel;

    public CrowdUsersSync(UserGroupManager userGroupManager, CrowdClient crowdClient, UserModel userModel, LoggerProvider loggerProvider) {
        this.userGroupManager = userGroupManager;
        this.loggerProvider = loggerProvider;
        this.crowdClient = crowdClient;
        this.userModel = userModel;
    }


    @Override
    public void run() {
        try {
            loggerProvider.getServerLogger().warn("--------------- SYNC ZE USERS NOW ------------------ ");
            syncUsers();
            loggerProvider.getServerLogger().warn("--------------- SYNC USERS COMPLETED WITH NO ERRORS------------------ ");
        } catch (Exception ex) {
            loggerProvider.getServerLogger().error("----------------- Massive problem when trying to sync the Crowd users.", ex);
        }
    }

    private void syncUsers() {
        final Map<String, User> originalBunchOfUsersBeforeTheUpdates =
                converter.allTeamCityUsersAsMap(userGroupManager.getAllUsersGroup().getAllUsers());
        updateUsers(originalBunchOfUsersBeforeTheUpdates);
    }

    private void updateUsers(Map<String, User> usersToUpdate) {
        loggerProvider.getServerLogger().info("Number of users to update: " + usersToUpdate.size());
        for (User user : usersToUpdate.values()) {
            try {
                loggerProvider.getServerLogger().info("Trying to update user " + user.getUsername());
                final com.atlassian.crowd.model.user.User crowdClientUser = crowdClient.getUser(user.getUsername());
                if (detailsAreDifferent(user, crowdClientUser)) {
                    updateUser(user, crowdClientUser);
                    loggerProvider.getServerLogger().info("User [" + user.getUsername() + "] details updated");
                }
            } catch (UserNotFoundException e) {
                loggerProvider.getServerLogger().info("Seems that user [" + user.getUsername() + "] doesn't exists in Crowd anymore. Need to remove from TC");
                userModel.removeUserAccount(user.getId());
            } catch (OperationFailedException e) {
                e.printStackTrace();
            } catch (ApplicationPermissionException e) {
                e.printStackTrace();
            } catch (InvalidAuthenticationException e) {
                e.printStackTrace();
            }
        }
    }

    private boolean detailsAreDifferent(User user, com.atlassian.crowd.model.user.User crowdClientUser) {
        return !StringUtils.equals(user.getName(), crowdClientUser.getName()) ||
                !StringUtils.equals(user.getEmail(), crowdClientUser.getEmailAddress());
    }

    private void updateUser(User user, com.atlassian.crowd.model.user.User crowdClientUser) {
        final SUser userAccount = userModel.findUserById(user.getId());
        userAccount.updateUserAccount(user.getUsername(), crowdClientUser.getName(), crowdClientUser.getEmailAddress());
    }
}
